/*
 * MDD Package
 * /projects/hsis/CVS/utilities/mdd/mdd.h,v
 * 
 * Author: Timothy Kam 
 * 1993/08/03 22:55:08
 * 1.2
 *
 * Copyright 1992 by the Regents of the University of California.
 *
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.  This software
 * is made available as is, with no warranties.
 */

/************************************************************************/
/***** RAMIN/TOM *****/
#ifndef MDD_DEFINED
#define MDD_DEFINED

#include "util.h"
#include "array.h"
#include "st.h"
#include "var_set.h"

#include "bdd.h" 

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif


/*    #include "bdd_int.h"   */

/************************************************************************/
#define str_len         10
#define MONITOR         0
#define VERBOSE         0
#define SOLUTION        0
#define BDD_SIZE        1
#define USE_ITE         1
#define BYPASS          1
/* code may not be fully debugged for interleaving */
#define INTERLEAVE      0

#define mdd_cofactor_min bdd_cofactor
#define mdd_and bdd_and
#define mdd_constant bdd_constant
#define mdd_dup bdd_dup
/* mdd_equal returns 1 if two MDD's are identical */
/* mdd_eq is a totally different function */
/* which returns an MDD instead */
#define mdd_equal bdd_equal
#define mdd_free bdd_free
#define mdd_is_tautology bdd_is_tautology
#define mdd_ite bdd_ite
/* mdd_lequal checks for implication, */
/* mdd_lequal(f,g,1,0) returns the value of (f => g') */
/* mdd_leq is a totally different function */
#define mdd_lequal bdd_leq
#define mdd_not bdd_not
#define mdd_one bdd_one
#define mdd_or bdd_or
#define mdd_size bdd_size
#define mdd_xor bdd_xor
#define mdd_xnor bdd_xnor
#define mdd_zero bdd_zero
#define mdd_EMPTY bdd_EMPTY
#define mdd_first_solution mdd_first_cube
#define mdd_next_solution mdd_next_cube

#define mdd_eq_c(m,a,b)  	mdd_func1c(m,a,b,eq2)
#define mdd_geq_c(m,a,b) 	mdd_func1c(m,a,b,geq2)
#define mdd_gt_c(m,a,b)  	mdd_func1c(m,a,b,gt2)
#define mdd_leq_c(m,a,b) 	mdd_func1c(m,a,b,leq2)
#define mdd_lt_c(m,a,b)  	mdd_func1c(m,a,b,lt2)
#define mdd_neq_c(m,a,b) 	mdd_func1c(m,a,b,neq2)

#define mdd_eq(m,a,b)  		mdd_func2(m,a,b,eq2)
#define mdd_geq(m,a,b) 		mdd_func2(m,a,b,geq2)
#define mdd_gt(m,a,b)  		mdd_func2(m,a,b,gt2)
#define mdd_leq(m,a,b) 		mdd_func2(m,a,b,leq2)
#define mdd_lt(m,a,b)  		mdd_func2(m,a,b,lt2)
#define mdd_neq(m,a,b) 		mdd_func2(m,a,b,neq2)
#define mdd_unary_minus(m,a,b) 	mdd_func2(m,a,b,unary_minus2)

#define mdd_eq_plus(m,a,b,c)	mdd_func3(m,a,b,c,eq_plus3)
#define mdd_geq_plus(m,a,b,c)	mdd_func3(m,a,b,c,geq_plus3)
#define mdd_gt_plus(m,a,b,c)	mdd_func3(m,a,b,c,gt_plus3)
#define mdd_leq_plus(m,a,b,c)	mdd_func3(m,a,b,c,leq_plus3)
#define mdd_lt_plus(m,a,b,c)	mdd_func3(m,a,b,c,lt_plus3)
#define mdd_neq_plus(m,a,b,c)	mdd_func3(m,a,b,c,neq_plus3)

#define mdd_eq_minus(m,a,b,c)	mdd_func3(m,a,b,c,eq_minus3)
#define mdd_geq_minus(m,a,b,c)	mdd_func3(m,a,b,c,geq_minus3)
#define mdd_gt_minus(m,a,b,c)	mdd_func3(m,a,b,c,gt_minus3)
#define mdd_leq_minus(m,a,b,c)	mdd_func3(m,a,b,c,leq_minus3)
#define mdd_lt_minus(m,a,b,c)	mdd_func3(m,a,b,c,lt_minus3)
#define mdd_neq_minus(m,a,b,c)	mdd_func3(m,a,b,c,neq_minus3)

#define mdd_eq_plus_c(m,a,b,c)	mdd_func2c(m,a,b,c,eq_plus3)
#define mdd_geq_plus_c(m,a,b,c)	mdd_func2c(m,a,b,c,geq_plus3)
#define mdd_gt_plus_c(m,a,b,c)	mdd_func2c(m,a,b,c,gt_plus3)
#define mdd_leq_plus_c(m,a,b,c)	mdd_func2c(m,a,b,c,leq_plus3)
#define mdd_lt_plus_c(m,a,b,c)	mdd_func2c(m,a,b,c,lt_plus3)
#define mdd_neq_plus_c(m,a,b,c)	mdd_func2c(m,a,b,c,neq_plus3)

/**** for backward compatibility only ****/
#define mdd_eq_g(m,a,b)  	mdd_func2(m,a,b,eq2)
#define mdd_geq_g(m,a,b) 	mdd_func2(m,a,b,geq2)
#define mdd_gt_g(m,a,b)  	mdd_func2(m,a,b,gt2)
#define mdd_leq_g(m,a,b) 	mdd_func2(m,a,b,leq2)
#define mdd_lt_g(m,a,b)  	mdd_func2(m,a,b,lt2)
#define mdd_neq_g(m,a,b) 	mdd_func2(m,a,b,neq2)
#define mdd_init_name(v,n,s) 	mdd_init(v,n,s)

/* mdd_int.h */
#define MDD_NOT(node)           ((bdd_node *) ((int) (node) ^ 01))
#define MDD_REGULAR(node)       ((bdd_node *) ((int) (node) & ~01))
#define MDD_IS_COMPLEMENT(node) ((int) (node) & 01)

#define MDD_ONE(bdd)            (bdd)->one
#define MDD_ZERO(bdd)           (MDD_NOT(MDD_ONE(bdd)))

#ifndef MAX
#define MAX(a,b) (a) > (b) ? (a) : (b)
#endif

#ifndef MIN
#define MIN(a,b) (a) < (b) ? (a) : (b)
#endif
/*  these definitions have been moved to bdd.h: */
/*
    typedef struct bdd_t mdd_t;
    typedef struct bdd_manager mdd_manager;
    typedef struct bdd_gen mdd_gen;
*/

/*	#define mdd_t bdd_t
	#define mdd_manager bdd_manager
*/
    typedef struct bdd_t mdd_t;
/* following line changed because of problem in compilation with cxx */
/*    typedef bdd_manager mdd_manager; */
    typedef bdd_manager mdd_manager;

struct mvar_type {
    int mvar_id;		/* mvar id */
    char *name;			/* name of mvar */
    int values;			/* no. of values mvar can take */
    int encode_length;		/* no. of binary variables, bvar's, */
				/* needed to encode mvar */
    int start_vertex;		/* first bvar id in bvar_list which */
				/* corresponds to mvar */
    int stride;			/* 0   : no interleaving */
				/* n>0 : interleave with n other mvar's */
    int *encoding;		/* internal use only */
};
typedef struct mvar_type mvar_type;

struct bvar_type {
    mdd_t *node;
    int mvar_id;
};
typedef struct bvar_type bvar_type;

struct mdd_hook_type {
    array_t *mvar_list;
    array_t *bvar_list;	
};
typedef struct mdd_hook_type mdd_hook_type;

struct mdd_gen {
    bdd_gen *bgen;
    bdd_gen_status status;
    array_t *cube;		/* array of literals {0,1,2} of all vars */
    array_t *minterm;		/* current minterm */
    array_t *var_list;		/* list of var id's */
    boolean out_of_range;
}; 
typedef struct mdd_gen mdd_gen;

#define foreach_mdd_minterm(fn, gen, minterm, var_list)\
  for((gen) = mdd_first_minterm(fn, &minterm, var_list);\
      ((gen)->status != bdd_EMPTY) ? TRUE: mdd_gen_free(gen);\
      (void) mdd_next_minterm(gen, &minterm))
                                                                      

extern mdd_hook_type mdd_hook;

/* external functions */
EXTERN mdd_manager *mdd_init ARGS((array_t*,array_t*,array_t*));
EXTERN unsigned int mdd_create_variables 
    ARGS((mdd_manager *,array_t *,array_t *, array_t *));
EXTERN unsigned int mdd_create_variables_after 
	ARGS((mdd_manager*, int, array_t*, array_t*, array_t* ));
EXTERN unsigned int mdd_create_variables_interleaved 
			ARGS((mdd_manager*, int, int, array_t*));
EXTERN void mdd_quit ARGS((mdd_manager *));
EXTERN mdd_t *mdd_case ARGS((mdd_manager *, int, array_t*));
EXTERN mdd_t *mdd_consensus ARGS((mdd_manager *,mdd_t*,array_t*));
EXTERN mdd_t *mdd_encode ARGS(());
EXTERN mdd_t *mdd_literal ARGS((mdd_manager*,int,array_t*));
EXTERN mdd_t *mdd_eq_s ARGS((mdd_manager*,int,int));
EXTERN mdd_t *mdd_neq_s ARGS((mdd_manager*,int,int));
EXTERN mdd_t *mdd_leq_s ARGS((mdd_manager*,int,int));
EXTERN mdd_t *mdd_geq_s ARGS((mdd_manager*,int,int));
EXTERN mdd_t *mdd_unary_minus_s ARGS(());
EXTERN mdd_t *mdd_lt_s ARGS((mdd_manager*,int,int));
EXTERN mdd_t *mdd_gt_s ARGS((mdd_manager*,int,int));
EXTERN void   mdd_search ARGS(());
EXTERN mdd_t *mdd_cofactor ARGS((mdd_manager *,mdd_t*,mdd_t*));
EXTERN mdd_t *mdd_smooth ARGS((mdd_manager *,mdd_t*,array_t*));
EXTERN mdd_t *mdd_and_smooth ARGS((mdd_manager *,mdd_t*,mdd_t*,array_t*));
EXTERN mdd_t *mdd_substitute ARGS((mdd_manager *,mdd_t*,array_t*,array_t*));
EXTERN array_t *mdd_get_support ARGS((mdd_manager *,mdd_t*));
EXTERN mdd_t *mdd_add_s ARGS((mdd_manager *,int,int,int));
EXTERN mdd_t *mdd_mod ARGS((mdd_manager *,int,int,int));
EXTERN mdd_t *mdd_interval ARGS(());
EXTERN double mdd_count_onset ARGS((mdd_manager*,mdd_t*,array_t*));
EXTERN mdd_t *mdd_cproject ARGS(());

/* mdd_iter.c */
EXTERN mdd_gen *mdd_first_minterm ARGS((mdd_t *f, array_t **minterm_p, 
	array_t *var_list));
EXTERN boolean mdd_next_minterm ARGS((mdd_gen *mgen, array_t **minterm_p));
EXTERN void mdd_print_array ARGS((array_t *array));
EXTERN int mdd_gen_free ARGS((mdd_gen *mgen));

EXTERN mdd_t *mdd_func1c ARGS((mdd_manager *mgr, int mvar1, int mvar2, 
	boolean (*func1c) (int, int)));

EXTERN mdd_t *mdd_func2 ARGS((mdd_manager *mgr, int mvar1, int mvar2, 
	boolean (*func2) (int, int)));
/* functions of 2 variables used by mdd_func2 ARGS(()) */
EXTERN boolean eq2 ARGS((int x, int y));
EXTERN boolean geq2 ARGS((int x, int y));
EXTERN boolean gt2 ARGS((int x, int y));
EXTERN boolean leq2 ARGS((int x, int y));
EXTERN boolean lt2 ARGS((int x, int y));
EXTERN boolean neq2 ARGS((int x, int y));
EXTERN boolean unary_minus2 ARGS((int x, int y));

EXTERN mdd_t *mdd_func2c ARGS((mdd_manager *mgr,int mvar1, int mvar2, int const,
	boolean (*func2c)(int, int, int)));
                                              
/* should be in ../test */
EXTERN mdd_t *mdd_func3 ARGS((mdd_manager *mgr, int mvar1, int mvar2, int mvar3,
	boolean (*func3)(int, int, int)));
/* functions of 3 variables used by mdd_func3 ARGS(()) */
EXTERN boolean eq_plus3 ARGS((int x, int y, int z));
EXTERN boolean geq_plus3 ARGS((int x, int y, int z));
EXTERN boolean gt_plus3 ARGS((int x, int y, int z));
EXTERN boolean leq_plus3 ARGS((int x, int y, int z));
EXTERN boolean lt_plus3 ARGS((int x, int y, int z));
EXTERN boolean neq_plus3 ARGS((int x, int y, int z));
/* actually functions below can be obtained from the set above */
/* by just suppling the negation of the constant value */
EXTERN boolean eq_minus3 ARGS((int x, int y, int z));
EXTERN boolean geq_minus3 ARGS((int x, int y, int z));
EXTERN boolean gt_minus3 ARGS((int x, int y, int z));
EXTERN boolean leq_minus3 ARGS((int x, int y, int z));
EXTERN boolean lt_minus3 ARGS((int x, int y, int z));
EXTERN boolean neq_minus3 ARGS((int x, int y, int z));

/* AML */
EXTERN array_t* mdd_ret_bvar_list ARGS((mdd_manager *man));
                                              
/* should be in ../test */
EXTERN mdd_t *mdd_process_constr ARGS(()); 
EXTERN mdd_t *mdd_fast_constr ARGS(()); 

/* should be in mdd_int.h */
EXTERN int toggle ARGS(());
EXTERN int no_bit_encode ARGS(());
EXTERN void print_strides ARGS(());
EXTERN void print_mvar_list ARGS(());
EXTERN void print_bdd_list_id ARGS(());
EXTERN void print_bvar_list_id ARGS(());
EXTERN void print_bdd ARGS(());
EXTERN mvar_type find_mvar_id ARGS(());
EXTERN void clear_all_marks ARGS(());
EXTERN void mdd_mark ARGS(());
EXTERN void mdd_unmark ARGS(());
EXTERN mvar_type find_mvar ARGS(());
EXTERN array_t *mdd_ret_mvar_list ARGS((mdd_manager *));
EXTERN int mdd_ret_bvar_id ARGS((mvar_type *,int));
EXTERN bvar_type mdd_ret_bvar ARGS((mvar_type *,int,array_t *));
EXTERN mdd_t *build_lt_c ARGS(());
EXTERN mdd_t *build_leq_c ARGS(());
EXTERN mdd_t *build_gt_c ARGS(());
EXTERN mdd_t *build_geq_c ARGS(());
EXTERN int num_bits ARGS(());
EXTERN int getbit ARGS(());
EXTERN array_t* mdd_ret_bvars_of_mvar ARGS(());
#endif
