#include "mdd.h"

/*
 * MDD Package
 * /projects/hsis/CVS/utilities/mdd/mdd_case.c,v
 *
 * Author: Timothy Kam
 * 1993/08/03 22:38:01
 * 1.2
 *
 * Copyright 1992 by the Regents of the University of California.
 *
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.  This software
 * is made available as is, with no warranties.
 */

/*
mdd_t *
mdd_case(mgr, mvar, child_list)
mdd_manager *mgr;
int mvar;
array_t *child_list;
{
    int i;
    mvar_type mv;
    mdd_t *mnode;
    array_t *mvar_list;
    array_t *intern_child_list;
    mdd_t *child, *intern_child;
    bdd_external_hooks *hook;

    hook =  bdd_get_external_hooks(mgr);
    mvar_list = ((mdd_hook_type *)(hook->mdd))->mvar_list;                     

    mv = array_fetch(mvar_type, mvar_list, mvar);
    
    if (mv.values != array_n(child_list)) 
	fail("mdd_case: mvar.values different from child_list\n");

    intern_child_list = array_alloc(mdd_t *, array_n(child_list));
    for (i=0; i<array_n(child_list); i++) {
	child = array_fetch(mdd_t *, child_list, i);
	intern_child = mdd_dup(child);
	array_insert(mdd_t *, intern_child_list, i, intern_child);
    }
	
    mnode = mdd_encode(mgr, 
	intern_child_list, 
	mv.start_vertex + mv.stride*(mv.encode_length-1),
	mv.stride);

    array_free(intern_child_list);
    return mnode;
}
*/

mdd_t *
mdd_case(mgr, mvar, child_list)
mdd_manager *mgr;
int mvar;
array_t *child_list;
{
    mvar_type mv;
    mdd_t *mnode;
    mdd_t *tmp;
    array_t *mvar_list;
    bdd_external_hooks *hook;

    hook =  bdd_get_external_hooks(mgr);
    mvar_list = ((mdd_hook_type *)(hook->mdd))->mvar_list;        


    mv = array_fetch(mvar_type, mvar_list, mvar);
    if (mv.values != array_n(child_list)) 
	fail("mdd_case: mvar.values different from child_list\n");
    if (mv.values == 1) {
	tmp = array_fetch(mdd_t *, child_list, 0);
	mnode = mdd_dup(tmp);
    }
    else {
        mnode = mdd_encode(mgr, child_list, 
	    mv.start_vertex + mv.stride*(mv.encode_length-1), mv.stride);
    }
    return mnode;
}

mdd_t *
mdd_encode(mgr, child_list, highest_vertex, stride)
mdd_manager *mgr;
array_t *child_list;
int highest_vertex;
int stride;
{
    array_t *new_child_list;
    register int i;
    register int child_count = 0;
    register int q = array_n(child_list);
    register bvar_type bv;
    mdd_t *f, *g, *h, *t;
    mdd_t *one, *zero;
    array_t *bvar_list;
    bdd_external_hooks *hook;

    hook =  bdd_get_external_hooks(mgr);
    bvar_list = ((mdd_hook_type *)(hook->mdd))->bvar_list;                                                                                

    one = mdd_one(mgr);    /* LEAK - fixed */
    zero = mdd_zero(mgr);  /* LEAK - fixed */


    /* print_bdd_list_id(child_list); */
    if (q == 1) {
        f =  array_fetch(mdd_t *, child_list, 0);
        mdd_free(one);
        mdd_free(zero);
        h = mdd_dup(f);
	if (!mdd_is_tautology(f,1) && !mdd_is_tautology(f,0)) mdd_free(f);
        return h;
    }
    new_child_list = array_alloc(mdd_t *, 0);
    for (i=0; i<(q/2); i++) {
	bv = array_fetch(bvar_type, bvar_list,  highest_vertex);
	f = mdd_dup(bv.node);  /* LEAK - fixed, but needs checking*/
        h = array_fetch(mdd_t *, child_list, child_count++);
        g = array_fetch(mdd_t *, child_list, child_count++);
#if USE_ITE
#if BYPASS
	/* bypasses cases 	*/
	/* 1  = ite(F,1,1)	*/
	/* 0  = ite(F,0,0)	*/
	/* F  = ite(F,1,0)	*/
	/* !F = ite(F,0,1)	*/
	/* G  = ite(F,G,G)	*/
    	if (mdd_is_tautology(g,0) && mdd_is_tautology(h,0)) {
	    array_insert_last(mdd_t *, new_child_list, zero);
	}
    	else if (mdd_is_tautology(g,0) && mdd_is_tautology(h,1)) {
	    t = mdd_not(f);
	    array_insert_last(mdd_t *, new_child_list, t);
	}
	else if (mdd_is_tautology(g,1) && mdd_is_tautology(h,1)) {
	    array_insert_last(mdd_t *, new_child_list, one);
	}
	else if (mdd_is_tautology(g,1) && mdd_is_tautology(h,0)) {
	    t = mdd_dup(f);  /* LEAK */
	    array_insert_last(mdd_t *, new_child_list, t);
            /*mdd_free(t);*/
	}
	else if (mdd_equal(f,g)) {
            t = mdd_dup(f);
	    array_insert_last(mdd_t *, new_child_list, t);
	}
	else {
    	    t = mdd_ite(f, g, h, 1, 1, 1); /* LEAK */
	    array_insert_last(mdd_t *, new_child_list, t);
            /* mdd_free(t); */
	}
        if (!mdd_is_tautology(g,1) && !mdd_is_tautology(g,0)) mdd_free(g);
        if (!mdd_is_tautology(h,1) && !mdd_is_tautology(h,0)) mdd_free(h);
#else
	t = mdd_ite(f, g, h, 1, 1, 1);
	if (!mdd_is_tautology(g,1) && !mdd_is_tautology(g,0)) mdd_free(g);
	if (!mdd_is_tautology(h,1) && !mdd_is_tautology(h,0)) mdd_free(h);
	array_insert_last(mdd_t *, new_child_list, t);
#endif
#else
	a1 = mdd_and(f,g,1,1);
	if (!mdd_is_tautology(g,1) && !mdd_is_tautology(g,0)) mdd_free(g);
	a2 = mdd_and(f,h,0,1);
	if (!mdd_is_tautology(h,1) && !mdd_is_tautology(h,0)) mdd_free(h);
	t = mdd_or(a1,a2,1,1);
	/* t = mdd_or(mdd_and(f,g,1,1), mdd_and(f,h,0,1), 1, 1); */
	if (!mdd_is_tautology(a1,1) && !mdd_is_tautology(a1,0)) mdd_free(a1);
	if (!mdd_is_tautology(a2,1) && !mdd_is_tautology(a2,0)) mdd_free(a2);
	array_insert_last(mdd_t *, new_child_list, t);
#endif
#if USE_ITE
        mdd_free(f);
#endif
    }
    if (q % 2) { 
	t = array_fetch(mdd_t *, child_list, child_count);
        array_insert_last(mdd_t *, new_child_list, t);
    }
    f =  mdd_encode(mgr, new_child_list, highest_vertex-stride, stride);
    array_free(new_child_list);
    mdd_free(one);
    mdd_free(zero);
    return f;
}
