#include "mdd.h"
#include "math.h"





/*
 * MDD Package
 * /projects/hsis/CVS/utilities/mdd/mdd_init.c,v
 *
 * Author: Timothy Kam
 * 1993/08/03 22:38:44
 * 1.2
 *
 * Copyright 1992 by the Regents of the University of California.
 *
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.  This software
 * is made available as is, with no warranties.
 */

static int
integer_get_num_of_digits(value)
int value;
{
    double x;
    int num_of_digits;

    if (value > 0) 	 x = log10((double) value); 
    else if (value == 0) x = 1;
    else 		 fail("mdd_init: internal error, value less than 0\n");
      
/*
 * Notice that floor(x) return a double, so it needs to be cast into an integer using
 * this following expression.
 */
    num_of_digits = floor(x) + 1;

    return num_of_digits;
}

mdd_manager *
mdd_init(mvar_values, mvar_names, mvar_strides)
array_t *mvar_values;
array_t *mvar_names;
array_t *mvar_strides;
{
    array_t *mvar_list;
    array_t *bvar_list;
    mvar_type mv;
    bvar_type bv;
    unsigned int i, j, no_mvars, bits, current_vertex;
    int vertex_sum;
    mdd_manager *mgr;
    char *mv_name;
    char *istr;
    char* name_str; 
    int stride_count, prev_stride;
    mdd_hook_type *mdd_hook;
    int n_bytes;
    bdd_external_hooks *hook;

    mdd_hook = ALLOC(mdd_hook_type, 1);

    /* global information about all mvar for mdd_manager */
    mvar_list = array_alloc(mvar_type, 0);

    /* global information about all bvar for mdd_manager */
    bvar_list = array_alloc(bvar_type, 0); 

    /* create the hook to the bdd_manager */
    mdd_hook->mvar_list = mvar_list;
    mdd_hook->bvar_list = bvar_list;

    /* if some array arguments are NIL */
    no_mvars = array_n(mvar_values);

    if (mvar_names == NIL(array_t)) {
	/* if no variable names are given, use generic naming */
	/* mv_0, mv_1, ... for variable 0, 1, ... */
	mvar_names = array_alloc(char *, 0);
	for (i=0; i<no_mvars; i++) {
            /* compose a name for the mvar */
            istr = ALLOC(char, integer_get_num_of_digits(i)+ 4);
            sprintf(istr, "mv_%d", (i + array_n(mvar_list)));
            array_insert_last(char *, mvar_names, istr);
/*
            (void) sprintf(istr, "%i", i);
	    mv_name = ALLOC(char, strlen(istr)+4);
            (void) strcpy(mv_name, mvar_name_prefix);
            (void) strcat(mv_name, istr);
            array_insert_last(char *, mvar_names, mv_name);
*/
	}
    }
    else {
        if (no_mvars != array_n(mvar_names)) 
	    fail("mdd_init: inconsistent size of mvar_names\n");
    }
    if (mvar_strides == NIL(array_t)) {
	/* if no strides are specified, the variables are not interleaved */
	/* i.e. mvar_strides = 1 */
	mvar_strides = array_alloc(int, 0);
	for (i=0; i<no_mvars; i++) {
	    array_insert_last(int, mvar_strides, 1);
        }
    }
    else {
	if (no_mvars != array_n(mvar_strides))
	    fail("mdd_init: inconsistent size of mvar_strides\n");
    }

    /* create mdd manager */      
    vertex_sum = 0;
    for (i=0; i<no_mvars; i++) {
        vertex_sum = vertex_sum + no_bit_encode(array_fetch(int,mvar_values,i));
    }
    mgr = bdd_start(vertex_sum);

    hook =  bdd_get_external_hooks(mgr);

    hook->mdd = (char *) mdd_hook;

    current_vertex = 0;
    stride_count = 1;
    prev_stride = -1;

#if VERBOSE
        printf("bdd variable insertion ordering: "); 
#endif

    /* mvar_values -> mvar_list */
    for (i=0; i<no_mvars; i++) {
	mv.mvar_id = i;

	/* compose a name for the mvar */
	name_str = array_fetch(char *, mvar_names, i);
	n_bytes = strlen (name_str) + 1;
        mv_name = ALLOC(char, n_bytes);
	strcpy(mv_name, name_str);
	mv.name = mv_name;

	/* register no. of values for the mvar */
	mv.values = array_fetch(int, mvar_values, i);

	/* length of ecoding bits for mvar */
        bits = no_bit_encode(mv.values);
	mv.encode_length = bits;

	/* register the starting bdd vertex no. for the mvar */
	mv.start_vertex = current_vertex;

	/* register strides */
	if (bits > 0)
	    mv.stride = array_fetch(int, mvar_strides, i);
	else
	    mv.stride = 1;

	/* create place-holder for bit-encoding of the mvar */
        mv.encoding = ALLOC(int, bits);

	/* create bdd variables and put them in bvar_list */
	for (j=0; j<bits; j++) {
	    bv.node = bdd_get_variable(mgr, current_vertex + mv.stride*j);
	    bv.mvar_id = i;

#if VERBOSE 
		printf("%d ", current_vertex + mv.stride*j);
#endif

	    array_insert(bvar_type, bvar_list, current_vertex + mv.stride*j, bv);
	}
	
	/* insert the completed mvar_type element to mvar_list */
        array_insert_last(mvar_type, mvar_list, mv);

	if ((prev_stride != -1 ) && (mv.stride != prev_stride)) {
	    printf("mdd_init: processing %s\n", mv.name);
	    fail("mdd_init: inconsistent found in mvar_stride\n");
	}

        /* register the stride for mvar interleaving */
        /* and update current bdd vertex count */
        if (stride_count == mv.stride) {
            stride_count = 1;
	    current_vertex = current_vertex + mv.stride*(bits-1) + 1;
	    prev_stride = -1;
	}
	else {
	    stride_count++;
	    current_vertex++;
	    prev_stride = mv.stride;
	}
    }

#if VERBOSE
	printf("\n");
#endif

    /* init all encodings to 2's */
    clear_all_marks(mgr);

#if MONITOR
    print_mvar_list(mgr);
    (void) printf("%d bdd variables created\n", array_n(bvar_list));
    print_bvar_list_id(mgr); 
#endif

   return mgr;
}


unsigned int
mdd_create_variables(mgr, mvar_values, mvar_names, mvar_strides)
mdd_manager *mgr;
array_t *mvar_values;
array_t *mvar_names;
array_t *mvar_strides;
{
    array_t *mvar_list;
    array_t *bvar_list;
    mvar_type mv;
    bvar_type bv;
    unsigned int i, j, no_mvars, bits, current_vertex, start_mvar;
    int vertex_sum;
    char *mv_name;
    char *istr;
    char* name_str; 
    int stride_count, prev_stride;
    int n_bytes;
    bdd_external_hooks *hook;
    bdd_t *temp;

    hook =  bdd_get_external_hooks(mgr);
    mvar_list = ((mdd_hook_type *)(hook->mdd))->mvar_list;                                                                                
    bvar_list = ((mdd_hook_type *)(hook->mdd))->bvar_list;                                                                                

 
    /* if some array arguments are NIL */
    no_mvars = array_n(mvar_values);

    if (mvar_names == NIL(array_t)) {
	/* if no variable names are given, use generic naming */
	/* mv_0, mv_1, ... for variable 0, 1, ... */
	mvar_names = array_alloc(char *, 0);
	for (i=0; i<no_mvars; i++) {
            /* compose a name for the mvar */
            istr = ALLOC(char, integer_get_num_of_digits(i + array_n(mvar_list))+ 4);
            sprintf(istr, "mv_%d", (i + array_n(mvar_list)));
            array_insert_last(char *, mvar_names, istr);
/* BUG ******
            (void) sprintf(istr, "%d", i + array_n(mvar_list));
	    mv_name = ALLOC(char, strlen(istr)+4);
            (void) strcpy(mv_name, mvar_name_prefix);
            (void) strcat(mv_name, istr);
            array_insert_last(char *, mvar_names, mv_name);
 */
	}
    }
    else {
        if (no_mvars != array_n(mvar_names)) 
	    fail("mdd_init: inconsistent size of mvar_names\n");
    }
    if (mvar_strides == NIL(array_t)) {
	/* if no strides are specified, the variables are not interleaved */
	/* i.e. mvar_strides = 1 */
	mvar_strides = array_alloc(char *, 0);
	for (i=0; i<no_mvars; i++) {
	    array_insert_last(int, mvar_strides, 1);
        }
    }
    else {
	if (no_mvars != array_n(mvar_strides))
	    fail("mdd_init: inconsistent size of mvar_strides\n");
    }

    vertex_sum = 0;
    for (i=0; i<no_mvars; i++) {
        vertex_sum = vertex_sum + no_bit_encode(array_fetch(int,mvar_values,i));
    }
    for (i=0; i<vertex_sum; i++) {
	temp =  bdd_create_variable(mgr);
	bdd_free(temp);
    }

    current_vertex = array_n(bvar_list);
    start_mvar = array_n(mvar_list);
    stride_count = 1;
    prev_stride = -1;

#if VERBOSE
        printf("bdd variable insertion ordering: "); 
#endif

    /* mvar_values -> mvar_list */
    for (i=0; i <no_mvars; i++) {
	mv.mvar_id = start_mvar + i;

	/* compose a name for the mvar */
	name_str = array_fetch(char *, mvar_names, i);
	n_bytes = strlen (name_str) + 1;
        mv_name = ALLOC(char, n_bytes);
	strcpy(mv_name, name_str);
	mv.name = mv_name;

	/* register no. of values for the mvar */
	mv.values = array_fetch(int, mvar_values, i);

	/* length of ecoding bits for mvar */
        bits = no_bit_encode(mv.values);
	mv.encode_length = bits;

	/* register the starting bdd vertex no. for the mvar */
	mv.start_vertex = current_vertex;

	/* register strides */
	if (bits > 0)
	    mv.stride = array_fetch(int, mvar_strides, i);
	else
	    mv.stride = 1;

	/* create place-holder for bit-encoding of the mvar */
        mv.encoding = ALLOC(int, bits);

	/* create bdd variables and put them in bvar_list */
	for (j=0; j<bits; j++) {
	    bv.node = bdd_get_variable(mgr, current_vertex + mv.stride*j);
	    bv.mvar_id = start_mvar+i;

#if VERBOSE 
		printf("%d ", current_vertex + mv.stride*j);
#endif

	    array_insert(bvar_type, bvar_list, current_vertex + mv.stride*j, bv);
	}
	
	/* insert the completed mvar_type element to mvar_list */
        array_insert_last(mvar_type, mvar_list, mv);

	if ((prev_stride != -1 ) && (mv.stride != prev_stride)) {
	    printf("mdd_init: processing %s\n", mv.name);
	    fail("mdd_init: inconsistent found in mvar_stride\n");
	}

        /* register the stride for mvar interleaving */
        /* and update current bdd vertex count */
        if (stride_count == mv.stride) {
            stride_count = 1;
	    current_vertex = current_vertex + mv.stride*(bits-1) + 1;
	    prev_stride = -1;
	}
	else {
	    stride_count++;
	    current_vertex++;
	    prev_stride = mv.stride;
	}
    }

#if VERBOSE
	printf("\n");
#endif

    /* init all encodings to 2's */
    clear_all_marks(mgr);

#if MONITOR
    print_mvar_list(mgr);
    (void) printf("%d bdd variables created\n", array_n(bvar_list));
    print_bvar_list_id(mgr); 
#endif

    return start_mvar;
}

unsigned int
mdd_create_variables_after(mgr, after_mvar_id,  mvar_values, mvar_names, mvar_strides)
mdd_manager *mgr;
int	after_mvar_id;
array_t *mvar_values;
array_t *mvar_names;
array_t *mvar_strides;
{
    array_t *mvar_list;
    array_t *bvar_list;
    mvar_type mv, after_mv;
    bvar_type bv;
    unsigned int i, j, no_mvars, bits, current_vertex, start_mvar;
    int vertex_sum;
    char *mv_name;
    char *istr;
    char* name_str; 
    int stride_count, prev_stride;
    int n_bytes, after_bv_id;
    bdd_external_hooks *hook;
    bdd_t *temp;
    bvar_type sec_bit_of_mv;
    int sec_bit_level, fol_id;

    hook =  bdd_get_external_hooks(mgr);
    mvar_list = ((mdd_hook_type *)(hook->mdd))->mvar_list;                                                                                
    bvar_list = ((mdd_hook_type *)(hook->mdd))->bvar_list;  

    after_mv = array_fetch( mvar_type, mvar_list, after_mvar_id );
    sec_bit_of_mv = array_fetch( bvar_type, bvar_list, after_mv.start_vertex+after_mv.stride );
    sec_bit_level = bdd_top_var_level( mgr, sec_bit_of_mv.node );

    /* first_bit_of_last_interleaved_mvar */
    fol_id = bdd_get_id_from_level( mgr, sec_bit_level - 1 );

    after_bv_id = fol_id + after_mv.stride * ( after_mv.encode_length - 1 );

    /* if some array arguments are NIL */
    no_mvars = array_n(mvar_values);

    if (mvar_names == NIL(array_t)) {
	/* if no variable names are given, use generic naming */
	/* mv_0, mv_1, ... for variable 0, 1, ... */
	mvar_names = array_alloc(char *, 0);
	for (i=0; i<no_mvars; i++) {
            /* compose a name for the mvar */
            istr = ALLOC(char, integer_get_num_of_digits(i + array_n(mvar_list))+ 4);
            sprintf(istr, "mv_%d", (i + array_n(mvar_list)));
            array_insert_last(char *, mvar_names, istr);
/* BUG ******
            (void) sprintf(istr, "%d", i + array_n(mvar_list));
	    mv_name = ALLOC(char, strlen(istr)+4);
            (void) strcpy(mv_name, mvar_name_prefix);
            (void) strcat(mv_name, istr);
            array_insert_last(char *, mvar_names, mv_name);
 */
	}
    }
    else {
        if (no_mvars != array_n(mvar_names)) 
	    fail("mdd_init: inconsistent size of mvar_names\n");
    }
    if (mvar_strides == NIL(array_t)) {
	/* if no strides are specified, the variables are not interleaved */
	/* i.e. mvar_strides = 1 */
	mvar_strides = array_alloc(char *, 0);
	for (i=0; i<no_mvars; i++) {
	    array_insert_last(int, mvar_strides, 1);
        }
    }
    else {
	if (no_mvars != array_n(mvar_strides))
	    fail("mdd_init: inconsistent size of mvar_strides\n");
    }

    vertex_sum = 0;
    for (i=0; i<no_mvars; i++) {
        vertex_sum = vertex_sum + no_bit_encode(array_fetch(int,mvar_values,i));
    }
    for (i=0; i<vertex_sum; i++) {
	temp =  bdd_create_variable_after(mgr, after_bv_id + i);
	bdd_free(temp);
    }

    current_vertex = array_n(bvar_list);
    start_mvar = array_n(mvar_list);
    stride_count = 1;
    prev_stride = -1;

#if VERBOSE
        printf("bdd variable insertion ordering: "); 
#endif

    /* mvar_values -> mvar_list */
    for (i=0; i <no_mvars; i++) {
	mv.mvar_id = start_mvar + i;

	/* compose a name for the mvar */
	name_str = array_fetch(char *, mvar_names, i);
	n_bytes = strlen (name_str) + 1;
        mv_name = ALLOC(char, n_bytes);
	strcpy(mv_name, name_str);
	mv.name = mv_name;

	/* register no. of values for the mvar */
	mv.values = array_fetch(int, mvar_values, i);

	/* length of ecoding bits for mvar */
        bits = no_bit_encode(mv.values);
	mv.encode_length = bits;

	/* register the starting bdd vertex no. for the mvar */
	mv.start_vertex = current_vertex;

	/* register strides */
	if (bits > 0)
	    mv.stride = array_fetch(int, mvar_strides, i);
	else
	    mv.stride = 1;

	/* create place-holder for bit-encoding of the mvar */
        mv.encoding = ALLOC(int, bits);

	/* create bdd variables and put them in bvar_list */
	for (j=0; j<bits; j++) {
	    bv.node = bdd_get_variable(mgr, current_vertex + mv.stride*j);
	    bv.mvar_id = start_mvar+i;

#if VERBOSE 
		printf("%d ", current_vertex + mv.stride*j);
#endif

	    array_insert(bvar_type, bvar_list, current_vertex + mv.stride*j, bv);
	}

	/* insert the completed mvar_type element to mvar_list */
        array_insert_last(mvar_type, mvar_list, mv);

	if ((prev_stride != -1 ) && (mv.stride != prev_stride)) {
	    printf("mdd_init: processing %s\n", mv.name);
	    fail("mdd_init: inconsistent found in mvar_stride\n");
	}

        /* register the stride for mvar interleaving */
        /* and update current bdd vertex count */
        if (stride_count == mv.stride) {
            stride_count = 1;
	    current_vertex = current_vertex + mv.stride*(bits-1) + 1;
	    prev_stride = -1;
	}
	else {
	    stride_count++;
	    current_vertex++;
	    prev_stride = mv.stride;
	}
    }

#if VERBOSE
	printf("\n");
#endif

    /* init all encodings to 2's */
    clear_all_marks(mgr);

#if MONITOR
    print_mvar_list(mgr);
    (void) printf("%d bdd variables created\n", array_n(bvar_list));
    print_bvar_list_id(mgr); 
#endif

    return start_mvar;
}

unsigned int
mdd_create_variables_interleaved(mgr, inter_var_id, no_mvars, mvar_names)
mdd_manager *mgr;
int	inter_var_id,
	no_mvars;
array_t *mvar_names;
{
    array_t *mvar_list;
    array_t *bvar_list;
    mvar_type mv, inter_var;
    bvar_type bv;
    unsigned int i, j, bits, current_vertex, start_mvar;
    char *mv_name;
    char *istr;
    char* name_str; 
    int stride_count, prev_stride;
    int n_bytes;
    bdd_external_hooks *hook;
    bdd_t *temp;


    hook =  bdd_get_external_hooks(mgr);
    mvar_list = ((mdd_hook_type *)(hook->mdd))->mvar_list;                                                                                
    bvar_list = ((mdd_hook_type *)(hook->mdd))->bvar_list;                                                                                

    inter_var = array_fetch( mvar_type, mvar_list, inter_var_id );


    /* if some array arguments are NIL */

    if (mvar_names == NIL(array_t)) {
	/* if no variable names are given, use generic naming */
	/* mv_0, mv_1, ... for variable 0, 1, ... */
	mvar_names = array_alloc(char *, 0);
	for (i=0; i<no_mvars; i++) {
            /* compose a name for the mvar */
            istr = ALLOC(char, integer_get_num_of_digits(i + array_n(mvar_list))+ 4);
            sprintf(istr, "mv_%d", (i + array_n(mvar_list)));
            array_insert_last(char *, mvar_names, istr);
/* BUG ******
            (void) sprintf(istr, "%d", i + array_n(mvar_list));
	    mv_name = ALLOC(char, strlen(istr)+4);
            (void) strcpy(mv_name, mvar_name_prefix);
            (void) strcat(mv_name, istr);
            array_insert_last(char *, mvar_names, mv_name);
 */
	}
    }


    for (j=0; j< inter_var.encode_length; j++) {
       for (i=0; i<no_mvars; i++) {
		temp = bdd_create_variable_after( mgr, inter_var.start_vertex + (inter_var.stride * j) );
		bdd_free( temp );
       }
    }


    current_vertex = array_n(bvar_list);
    start_mvar = array_n(mvar_list);
    stride_count = 1;
    prev_stride = -1;

#if VERBOSE
        printf("bdd variable insertion ordering: "); 
#endif

    /* mvar_values -> mvar_list */
    for (i=0; i <no_mvars; i++) {
	mv.mvar_id = start_mvar + i;

	/* compose a name for the mvar */
	name_str = array_fetch(char *, mvar_names, i);
	n_bytes = strlen (name_str) + 1;
        mv_name = ALLOC(char, n_bytes);
	strcpy(mv_name, name_str);
	mv.name = mv_name;

	/* register no. of values for the mvar */
	mv.values = inter_var.values;

	/* length of ecoding bits for mvar */
        bits = no_bit_encode(mv.values);
	mv.encode_length = bits;

	/* register the starting bdd vertex no. for the mvar */
	mv.start_vertex = current_vertex;

	/* register strides */
	mv.stride = no_mvars;

	/* create place-holder for bit-encoding of the mvar */
        mv.encoding = ALLOC(int, bits);

	/* create bdd variables and put them in bvar_list */
	for (j=0; j<bits; j++) {
	    bv.node = bdd_get_variable(mgr, current_vertex + mv.stride*j);
	    bv.mvar_id = start_mvar+i;

#if VERBOSE 
		printf("%d ", current_vertex + mv.stride*j);
#endif

	    array_insert(bvar_type, bvar_list, current_vertex + mv.stride*j, bv);
	}
	
	/* insert the completed mvar_type element to mvar_list */
        array_insert_last(mvar_type, mvar_list, mv);

	if ((prev_stride != -1 ) && (mv.stride != prev_stride)) {
	    printf("mdd_init: processing %s\n", mv.name);
	    fail("mdd_init: inconsistent found in mvar_stride\n");
	}

        /* register the stride for mvar interleaving */
        /* and update current bdd vertex count */
        if (stride_count == mv.stride) {
            stride_count = 1;
	    current_vertex = current_vertex + mv.stride*(bits-1) + 1;
	    prev_stride = -1;
	}
	else {
	    stride_count++;
	    current_vertex++;
	    prev_stride = mv.stride;
	}
    }

#if VERBOSE
	printf("\n");
#endif

    /* init all encodings to 2's */
    clear_all_marks(mgr);

#if MONITOR
    print_mvar_list(mgr);
    (void) printf("%d bdd variables created\n", array_n(bvar_list));
    print_bvar_list_id(mgr); 
#endif

    return start_mvar;
}
