#include "mdd.h"

mdd_t *
build_leq_c(mgr, mvar_id, c) /* var <= c */
mdd_manager *mgr;
int mvar_id,
    c;
{
	return build_lt_c(mgr, mvar_id, c+1);
}
mdd_t * 
build_lt_c(mgr, mvar_id, c)
mdd_manager *mgr;
int mvar_id,
    c;
{
	mdd_t *A, *one, *zero;
	array_t *bvar_list, *mvar_list;
	mvar_type a_mv;
	int i;
	bvar_type bit_i;

	mvar_list = mdd_ret_mvar_list(mgr);
	bvar_list = mdd_ret_bvar_list(mgr);

	one = mdd_one(mgr);
	zero = mdd_zero(mgr);

	a_mv = array_fetch( mvar_type, mvar_list, mvar_id);

	if ( a_mv.values <= c ) {
		mdd_free(zero);
		return one;
	}

	A = mdd_zero(mgr);

	for(i=1; i <= a_mv.encode_length; i++){
		bit_i = array_fetch( bvar_type, bvar_list, 
					(a_mv.start_vertex + (a_mv.encode_length - i)* a_mv.stride )   );

		if ( getbit(c,i-1) == 0 ) 
			A = mdd_ite(bit_i.node, zero,   A, 1, 1, 1);
		else
			A = mdd_ite(bit_i.node,    A, one, 1, 1, 1);
		
	}

	mdd_free(one);
	mdd_free(zero);
		
	return A;
}

mdd_t * 
build_geq_c(mgr, mvar_id, c)
mdd_manager *mgr;
int mvar_id,
    c;
{
	mdd_t *A, *one, *zero;
	array_t *bvar_list, *mvar_list;
	mvar_type a_mv;
	int i;
	bvar_type bit_i;

	mvar_list = mdd_ret_mvar_list(mgr);
	bvar_list = mdd_ret_bvar_list(mgr);

	one = mdd_one(mgr);
	zero = mdd_zero(mgr);

	a_mv = array_fetch( mvar_type, mvar_list, mvar_id);

	if ( a_mv.values <= c ) { 
		mdd_free( one );
		return zero;
	}

	A = mdd_one(mgr);

	for(i=1; i <= a_mv.encode_length; i++){
		bit_i = array_fetch( bvar_type, bvar_list, 
					(a_mv.start_vertex + (a_mv.encode_length - i)* a_mv.stride )   );

		if ( getbit(c,i-1) == 0 ) 
			A = mdd_ite(bit_i.node, one,    A, 1, 1, 1);
		else
			A = mdd_ite(bit_i.node,   A, zero, 1, 1, 1);
		
	}

	mdd_free(one);
	mdd_free(zero);
		
	return A;
} 

mdd_t *
build_gt_c(mgr, mvar_id, c)
mdd_manager *mgr;
int mvar_id,
    c;
{
	return build_geq_c(mgr, mvar_id, c+1);
}

mdd_t *
mdd_interval(mgr, mvar_id, low, high)  /* low <= var <= high */
mdd_manager *mgr;
int mvar_id,
    low,
    high;
{
	mdd_t *HIGH_MDD, *LOW_MDD, *result;

	LOW_MDD = build_geq_c(mgr,mvar_id,low);
	HIGH_MDD = build_leq_c(mgr,mvar_id,high);

	result = mdd_and(HIGH_MDD, LOW_MDD, 1, 1);

	mdd_free(LOW_MDD);
	mdd_free(HIGH_MDD);

	return result;
}
