#include <stdio.h>
#include <math.h>
#include "mdd.h"

/*
 * MDD Package
 * /projects/hsis/CVS/utilities/mdd/mdd_util.c,v
 *
 * Author: Timothy Kam
 * 1993/08/03 22:39:43
 * 1.2
 *
 * Copyright 1992 by the Regents of the University of California.
 *
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.  This software
 * is made available as is, with no warranties.
 */

/****************** Definitions taken from Ramin            *************/
	double		mdd_count_onset( );
	array_t*	mddRetBinVars( );
	array_t*	mddRetArrBinVars( );

static	bdd_t*		mddRetOnvalBdd( );
static	bdd_t*		mddIntRetOnvalBdd( );
	void		mddFreeBddArr( );
        mdd_t*          mdd_cproject( );

/************************************************************************/
extern	double		bdd_count_onset( );
extern	mdd_t*		mdd_forall( );
	
/************************************************************************/
#define		mddGetVarById( mgr, id )	\
    array_fetch(mvar_type, mdd_ret_mvar_list((mgr)),(id))

int
toggle(x)
int x;
{
    if (x == 0) return 1;
    else {
	if (x == 1) return 0;
	else {
	    fail("toggle: invalid boolean value\n");
	    return -1;
	}
    }
}

int
no_bit_encode(n)
int n;
{
    int i = 0;
    int j = 1;

    if (n < 2) return 1; /* Takes care of constants, i.e. mvars with just one value */

    while (j < n) {
	j = j * 2;
	i++;
    }
    return i;
}

void
print_mvar_list(mgr)
mdd_manager *mgr;
{
    mvar_type mv;
    int i;
    int no_mvar;
    array_t *mvar_list;
    bdd_external_hooks *hook;

    hook =  bdd_get_external_hooks(mgr);
    mvar_list = ((mdd_hook_type *)(hook->mdd))->mvar_list;      
    no_mvar = array_n(mvar_list);
    printf("print_mvar_list:\n");
    printf("id\tname\tvalues\tbits\tstride\tstart_vertex\n");
    for (i=0; i<no_mvar; i++) {
        mv = array_fetch(mvar_type, mvar_list, i);
        (void) printf("%d\t%s\t%d\t%d\t%d\t%d\n", 
		mv.mvar_id, mv.name, mv.values, 
		mv.encode_length, mv.stride, mv.start_vertex);
    }
}

void
print_strides(mvar_strides)
array_t *mvar_strides;
{
    int i, s;

    (void) printf("mvar_strides: ");
    for (i=0; i<array_n(mvar_strides); i++) {
        s = array_fetch(int, mvar_strides, i);
        (void) printf("%d ", s);
    }
    (void) printf("\n");
}

void
print_bdd_list_id(bdd_list)
array_t *bdd_list;
{
    bdd_t *b;
    int i, is_complemented;

    (void) printf("bdd_list id's: ");
    for (i=0; i<array_n(bdd_list); i++) {
        b = array_fetch(bdd_t *, bdd_list, i);
        (void)bdd_get_node(b, &is_complemented);
        if (is_complemented) (void) printf("!");
        (void) printf("%d ", bdd_top_var_id(b));
    }
    (void) printf("\n");
}

void
print_bvar_list_id(mgr)
mdd_manager *mgr;
{
    bvar_type bv;
    int i, is_complemented;
    array_t *bvar_list;
    bdd_external_hooks *hook;

    hook =  bdd_get_external_hooks(mgr);
    bvar_list = ((mdd_hook_type *)(hook->mdd))->bvar_list;
    
    (void) printf("bvar_list id's: ");
    for (i=0; i<array_n(bvar_list); i++) {
        bv = array_fetch(bvar_type, bvar_list, i);
        (void)bdd_get_node(bv.node,&is_complemented);
        if (is_complemented) (void) printf("!");
	(void) printf("%d ", bdd_top_var_id(bv.node)); /* Temporary fix for Arlindo */
    }
    (void) printf("\n");
}

void
print_bdd(mgr, top)
bdd_manager *mgr;
bdd_t *top;  /****  was: bdd_node *bnode; --- changed by Serdar ****/
{

    int is_complemented;
    bdd_t *child, *top_uncomp;

    if (bdd_is_tautology(top,1)) {
	(void) printf("ONE ");
        return;
    }
    if (bdd_is_tautology(top,0)) {
	(void) printf("ZERO ");
        return;
    }
    (void)bdd_get_node(top, &is_complemented);
    if (is_complemented != 0) (void) printf("!");
    (void) printf("%d ", bdd_top_var_id(top));
    (void) printf("v ");
    (void) printf("< ");

    if (is_complemented) top_uncomp = bdd_not(top);
    else top_uncomp = mdd_dup(top);

    child = bdd_then(top);

    print_bdd(mgr, child);
    (void) printf("> ");

    mdd_free(child);
    child = bdd_else(top);

    print_bdd(mgr, child);
    (void) printf("^ ");
    
    mdd_free(top_uncomp);
    mdd_free(child);

    return;
}

mvar_type 
find_mvar_id(mgr, id)
mdd_manager *mgr;
unsigned short id;
{
    mvar_type mv;
    bvar_type bv;
    array_t *mvar_list;
    array_t *bvar_list;
    bdd_external_hooks *hook;

    hook =  bdd_get_external_hooks(mgr);
    mvar_list = ((mdd_hook_type *)(hook->mdd))->mvar_list;                                                                                
    bvar_list = ((mdd_hook_type *)(hook->mdd))->bvar_list;          

    if (id >= array_n(bvar_list))
    	fail("find_mvar_id: invalid parameter range for id\n");
    bv = array_fetch(bvar_type, bvar_list, id);
    if ((bv.mvar_id < 0) || (bv.mvar_id >= array_n(mvar_list)))
    	fail("find_mvar_id: bvar contains invalid mvar_id\n");
    mv = array_fetch(mvar_type, mvar_list, bv.mvar_id);
    return mv;
}

void
clear_all_marks(mgr)
mdd_manager *mgr;
{
    int i, j;
    mvar_type mv;
    array_t *mvar_list;
    bdd_external_hooks *hook;

    hook =  bdd_get_external_hooks(mgr);
    mvar_list = ((mdd_hook_type *)(hook->mdd))->mvar_list;                                                                                


    for (i=0; i<array_n(mvar_list); i++) {
	mv = array_fetch(mvar_type, mvar_list, i);
	for (j=0; j<mv.encode_length; j++)
	    mv.encoding[j] = 2;
    }
}

void
mdd_mark(mgr, top, phase)
mdd_manager *mgr;
bdd_t *top; /**** was bdd_node *bnode; --- changed by Serdar ***/
int phase;
{
    int bit_position, dummy;
    mvar_type mv;
    

    (void)bdd_get_node(top,&dummy);

    mv = find_mvar_id(mgr, bdd_top_var_id(top));
    bit_position = (bdd_top_var_id(top) - mv.start_vertex) / mv.stride;
    if (((bdd_top_var_id(top) - mv.start_vertex) % mv.stride) != 0)
	fail("mdd_mark: interleaving error\n");
    mv.encoding[bit_position] = phase;
/*    mdd_free(top_node); Temoprary fix for Arlindo */ 
}

void
mdd_unmark(mgr, top)
mdd_manager *mgr;
bdd_t *top; /**** was bdd_node *bnode; --- changed by Serdar ***/
{
    int bit_position, dummy;
    mvar_type mv;

    (void)bdd_get_node(top,&dummy);


    mv = find_mvar_id(mgr, bdd_top_var_id(top));
    bit_position = (bdd_top_var_id(top) - mv.start_vertex) / mv.stride;
    if (((bdd_top_var_id(top) - mv.start_vertex) % mv.stride) != 0)
	fail("mdd_unmark: interleaving error\n");
    mv.encoding[bit_position] = 2;
}

mvar_type 
find_mvar(mgr, name)
mdd_manager *mgr;
char *name;
{
    int i;
    mvar_type mv;
    array_t *mvar_list;
    bdd_external_hooks *hook;

    hook =  bdd_get_external_hooks(mgr);
    mvar_list = ((mdd_hook_type *)(hook->mdd))->mvar_list;    

    for (i=0; i<array_n(mvar_list); i++) {
	mv = array_fetch(mvar_type, mvar_list, i);
        if (strcmp(mv.name, name) == 0) return mv;
    }
    fail("find_mvar: cannot find name in mvar_list\n");
    return mv;
}

array_t *
mdd_ret_mvar_list(mgr)
mdd_manager *mgr;
{
    bdd_external_hooks *hook;    
    array_t *mvar_list;

    hook =  bdd_get_external_hooks(mgr);
    mvar_list = ((mdd_hook_type *)(hook->mdd))->mvar_list;

    return mvar_list;
}

void
mdd_set_mvar_list(mgr, mvar_list)
mdd_manager *mgr;
array_t *mvar_list;
{
    bdd_external_hooks *hook;    

    hook =  bdd_get_external_hooks(mgr);
    ((mdd_hook_type *)(hook->mdd))->mvar_list = mvar_list;
}

array_t *
mdd_ret_bvar_list(mgr)
mdd_manager *mgr;
{
    bdd_external_hooks *hook;    
    array_t *bvar_list;

    hook =  bdd_get_external_hooks(mgr);
    bvar_list = ((mdd_hook_type *)(hook->mdd))->bvar_list;

    return bvar_list;
}

/************************************************************************/
/* This pacakege implements counting the num of onset points of an Mdd.	*/
/* Author:	Ramin Hojati		 November 1991.			*/

/************************************************************************/
/* Given an Mdd, returns the num of onset points.  By construction of	*/
/* Mdd's, some points not in the range of Mdd vars may be inculded 	*/
/* in the onset. These fake points must first be removed.		*/

double
mdd_count_onset( mddMgr, aMdd, mddIdArr )
	mdd_manager	*mddMgr;
	mdd_t		*aMdd;
	array_t		*mddIdArr;
{	
	bdd_t		*onvalBdd, *aOnvalBdd, *onsetBdd, *tmpBdd;
	double		onsetNum;
	array_t		*bddVarArr;
	int		i, arrSize, mddId;

	arrSize = array_n( mddIdArr );
	onvalBdd = bdd_one( mddMgr );

	for ( i = 0 ; i < arrSize ; i++ ) {
	    mddId = array_fetch( int, mddIdArr, i );
	    aOnvalBdd = mddRetOnvalBdd( mddMgr, mddId );

	    tmpBdd = bdd_and( onvalBdd, aOnvalBdd, 1, 1 );
	    bdd_free( onvalBdd );
	    bdd_free( aOnvalBdd );
	    onvalBdd = tmpBdd;
	}
	onsetBdd = bdd_and( onvalBdd, aMdd, 1, 1 );
	bdd_free( onvalBdd );

	bddVarArr = mddRetArrBinVars( mddMgr, mddIdArr );
	onsetNum = bdd_count_onset( onsetBdd, bddVarArr );
	bdd_free( onsetBdd );
	mddFreeBddArr( bddVarArr );
	return( onsetNum );
}		/* mdd_count_onset */

/************************************************************************/
array_t*
mddRetBinVars( mddMgr, mddId )
	mdd_manager	*mddMgr;
	int		mddId;
{
	array_t		*bddVarArr;
	mvar_type	mVar;
	unsigned int	i, lastIndex;

	mVar = mddGetVarById( mddMgr, mddId );
	bddVarArr = array_alloc( bdd_t *, mVar.encode_length );	

	lastIndex = mVar.start_vertex + mVar.stride * mVar.encode_length;
        for (i = mVar.start_vertex ; i <  lastIndex ; i += mVar.stride ) {
	    array_insert_last( bdd_t *, bddVarArr, 
			       bdd_get_variable( mddMgr, i ));
	}
	return( bddVarArr );
}		/* mddRetBinVars */

/************************************************************************/
/* Returns an array of binary vars (bdd_t *'s), given an array of 	*/
/* Mdd id's.								*/

array_t*
mddRetArrBinVars( mddMgr, mddIdArr )
	mdd_manager	*mddMgr;
	array_t		*mddIdArr;
{
	array_t		*bddVarArr;
	unsigned int	j, i, arrSize, lastIndex;
	mvar_type	mVar;

	bddVarArr = array_alloc( bdd_t *, 3 * array_n( mddIdArr ));	

	arrSize = array_n( mddIdArr );
	for ( i = 0 ; i < arrSize ; i++ ) {
	    mVar = mddGetVarById( mddMgr, array_fetch( int, mddIdArr, i ));
	    lastIndex = mVar.start_vertex + mVar.stride * mVar.encode_length;
            for ( j = mVar.start_vertex ; j < lastIndex ; j += mVar.stride ) {
		array_insert_last( bdd_t *, bddVarArr, 
				   bdd_get_variable( mddMgr, j ));
	    }
	}
	return( bddVarArr );
}		/* mddRetArrBinVars */

/************************************************************************/
static	bdd_t*
mddRetOnvalBdd( mddMgr, mddId )
	mdd_manager	*mddMgr;
	int		mddId;
{
	bdd_t		*onvalBdd;
	mvar_type	mVar;
	int		valNum, high;
	array_t		*bddVarArr;	
	
	mVar = mddGetVarById( mddMgr, mddId );
	valNum = mVar.values;
	high = (int) pow( (double) 2, (double) mVar.encode_length ); 
	assert( (valNum <= high) && (valNum > high/2) );
	if ( valNum == high )
	    onvalBdd = bdd_one( mddMgr );
	else {
	    bddVarArr = mddRetBinVars( mddMgr, mddId );
	    onvalBdd = mddIntRetOnvalBdd( mddMgr, valNum, 0, high, 
					  0, bddVarArr );
	    mddFreeBddArr( bddVarArr );
	}
	return( onvalBdd );
}		/* mddRetOnvalBdd */	

/************************************************************************/
static	bdd_t*
mddIntRetOnvalBdd( mddMgr, valNum, low, hi, level, bddVarArr )
	mdd_manager	*mddMgr;
	int		valNum, low, hi, level;
	array_t		*bddVarArr;
{
	int		mid;
	bdd_t		*onvalBdd, *curVar, *recBdd, *tmpBdd;

	mid = (low + hi) / 2;
		curVar = array_fetch( bdd_t *, bddVarArr, level );

	if 	( valNum > mid ) {
	    recBdd = mddIntRetOnvalBdd( mddMgr, valNum, mid, hi, 
					level+1, bddVarArr );
	    tmpBdd = bdd_and( recBdd, curVar, 1, 1 );
	    bdd_free( recBdd );
	    onvalBdd = bdd_or( tmpBdd, curVar, 1, 0 );
	    bdd_free( tmpBdd );
	}
	else if ( valNum < mid ) {
	    recBdd = mddIntRetOnvalBdd( mddMgr, valNum, low, mid, 
					level+1, bddVarArr );
	    onvalBdd = bdd_and( recBdd, curVar, 1, 0 );
	    bdd_free( recBdd );
	}
	else if ( valNum == mid ) 
	    onvalBdd = bdd_not( curVar );
	return( onvalBdd );
}		/* mddIntRetOnvalBdd */

/************************************************************************/
/* Given an array of bdd nodes, frees the array.			*/

void
mddFreeBddArr( bddArr )
	array_t	*bddArr;
{
	int	i, arrSize;

	arrSize = array_n( bddArr );
	for ( i = 0 ; i < arrSize ; i++ ) 
	    bdd_free( array_fetch( bdd_t *, bddArr, i ) );
	array_free( bddArr );
}		/* mddFreeBddArr */

/***************************************************************************/

mdd_t*
mdd_cproject( mddMgr, aMdd, nsIds )
        mdd_manager     *mddMgr;
        mdd_t           *aMdd;
        array_t         *nsIds;
{
        mdd_t           *res;
        array_t         *bddVarArr;

        bddVarArr = mddRetArrBinVars( mddMgr, nsIds );
        res = bdd_cproject( aMdd, bddVarArr );
        mddFreeBddArr( bddVarArr );
        return( res );
}               /* mdd_cproject */

bvar_type
mdd_ret_bvar( mvar_ptr, index, bvar_list )
mvar_type 	*mvar_ptr;
int		index;
array_t		*bvar_list;
{
	bvar_type	result;

	result = array_fetch( bvar_type, bvar_list, mvar_ptr->start_vertex + (index * mvar_ptr->stride) );

	return result;


}

int
mdd_ret_bvar_id( mvar_ptr, index )
mvar_type 	*mvar_ptr;
int		index;
{

	return ( mvar_ptr->start_vertex + (index * mvar_ptr->stride) );

}

array_t  *
mdd_ret_bvars_of_mvar( mvar_ptr )
mvar_type	*mvar_ptr;
{
	array_t *ids_arr = array_alloc( int, 0 );
	int i,		index = mvar_ptr->start_vertex;

	for(i=0; i<mvar_ptr->encode_length; i++) {
		array_insert_last( int, ids_arr, index );
		index += mvar_ptr->stride;
	}

	return ids_arr;
}











